import sys
from collections import defaultdict, OrderedDict
from natsort import natsorted
from itertools import product
import numpy as np
from .utils import zopen
from .gff import gff2dict
from .interlap import InterLap
from .consensus import getAED


def compare(args):
    compareAnnotations(args.reference, args.query, args.fasta, output=args.out)


def gff2interlap(input, fasta):
    """
    function to parse GFF3 file, construct scaffold/gene interlap dictionary and funannotate standard annotation dictionary
    """
    inter = defaultdict(InterLap)
    Genes = gff2dict(input, fasta)
    for k, v in natsorted(list(Genes.items())):
        inter[v["contig"]].add((v["location"][0], v["location"][1], k))
    return inter, Genes


def countFeatures(input):
    # given funannotate dictionary, count up some general features
    mRNAg, mRNAt, tRNAg, tRNAt = (0,) * 4
    for k, v in natsorted(list(input.items())):
        if v["type"] == "mRNA":
            mRNAg += 1
            mRNAt += len(v["ids"])
        elif v["type"] == "tRNA":
            tRNAg += 1
            tRNAt += len(v["ids"])
    return len(input), mRNAg, mRNAt, tRNAg, tRNAt


def pairwiseAED(query, reference):
    """
    takes a multiple transcripts and sums AED from lowest pairwise comparison and then calculates
    the average based on number of transcripts in the query
    """
    AEDsum = []
    pAED = [float(getAED(a, b)) for a, b in product(query, reference)]
    # split into parts to get lowest AED
    splitAED = [pAED[i : i + len(query)] for i in range(0, len(pAED), len(query))]
    for pair in splitAED:
        AEDsum.append(min(pair))
    AEDavg = sum(AEDsum) / len(query)
    return "{:.3f}".format(AEDavg)


def compareAnnotations(old, new, fasta, output=False):
    """
    function takes two GFF annotated genomes and compares gene models
    output is a tsv file for each locus and a description of what is different
    can handle multiple transcripts per locus
    """
    if output:
        out = zopen(output, mode="w")
    else:
        out = sys.stdout
    result = {}
    global no_change, identicalCDS, refUnique, queryUnique
    no_change, identicalCDS, refUnique, queryUnique, totalmatches, totallength = (
        0,
    ) * 6
    oldInter, oldGenes = gff2interlap(old, fasta)
    NumOldLoci, NumOldGenes, NumOldmRNA, NumOldtRNALoci, NumOldtRNA = countFeatures(
        oldGenes
    )
    sys.stderr.write(
        "Reference: {} contigs {} genes\n".format(len(oldInter), len(oldGenes))
    )
    sys.stderr.write(
        "{}\n".format([NumOldLoci, NumOldGenes, NumOldmRNA, NumOldtRNALoci, NumOldtRNA])
    )
    newInter, newGenes = gff2interlap(new, fasta)
    sys.stderr.write(
        "Query: {} contigs {} genes\n".format(len(newInter), len(newGenes))
    )
    NumNewLoci, NumNewGenes, NumNewmRNA, NumNewtRNALoci, NumNewtRNA = countFeatures(
        newGenes
    )
    sys.stderr.write(
        "{}\n".format([NumNewLoci, NumNewGenes, NumNewmRNA, NumNewtRNALoci, NumNewtRNA])
    )

    # now go through the updated annotation, comparing to old annot
    ref_seen = set()
    for contig in newInter:
        for gene in newInter[contig]:
            # means this is a new model, so add it
            hitList = list(oldInter[contig].find((gene[0], gene[1])))
            if len(hitList) < 1:
                result[gene[2]] = {
                    "contig": newGenes[gene[2]]["contig"],
                    "location": newGenes[gene[2]]["location"],
                    "ref_type": None,
                    "ref_location": None,
                    "query_location": newGenes[gene[2]]["location"],
                    "query_id": gene[2],
                    "query_type": newGenes[gene[2]]["mRNA"],
                    "ref_id": None,
                    "cdsAED": "1.000",
                    "exonAED": "1.000",
                    "ref_transcripts": 0,
                    "query_transcripts": len(newGenes[gene[2]]["ids"]),
                    "ref_strand": None,
                    "query_strand": newGenes[gene[2]]["strand"],
                }
                continue
            # there might be some overlapping transcripts, so get best hit?
            hit = []
            # get best hit
            for z in hitList:
                diffs = np.subtract((gene[0], gene[1]), (z[0], z[1]))
                totaldiffs = abs(diffs[0]) + abs(diffs[1])
                hit.append((totaldiffs, z[2]))
            besthit = min(hit)

            # get the old annotation
            hitInfo = oldGenes.get(besthit[1])
            ref_seen.add(besthit[1])

            # calculate AED
            exonAED = pairwiseAED(newGenes[gene[2]]["mRNA"], hitInfo["mRNA"])
            if "mRNA" in newGenes[gene[2]]["type"] and "mRNA" in hitInfo["type"]:
                cdsAED = pairwiseAED(newGenes[gene[2]]["CDS"], hitInfo["CDS"])
            else:
                cdsAED = "0.000"

            if not besthit[1] in result:
                result[besthit[1]] = {
                    "contig": newGenes[gene[2]]["contig"],
                    "location": hitInfo["location"],
                    "ref_type": hitInfo["type"],
                    "ref_location": hitInfo["location"],
                    "query_location": newGenes[gene[2]]["location"],
                    "query_id": gene[2],
                    "query_type": newGenes[gene[2]]["type"],
                    "cdsAED": cdsAED,
                    "exonAED": exonAED,
                    "ref_transcripts": len(hitInfo["ids"]),
                    "query_transcripts": len(newGenes[gene[2]]["ids"]),
                    "ref_strand": hitInfo["strand"],
                    "query_strand": newGenes[gene[2]]["strand"],
                    "ref_id": besthit[1],
                }
            # get some summary stats as you loop through
            if float(exonAED) == 0 and float(cdsAED) == 0:
                no_change += 1
            elif float(cdsAED) == 0:
                identicalCDS += 1
    # now add old genes that did not have overlaps
    for contig in oldInter:
        for gene in oldInter[contig]:
            if not gene[2] in ref_seen:
                result[gene[2]] = {
                    "contig": oldGenes[gene[2]]["contig"],
                    "location": oldGenes[gene[2]]["location"],
                    "ref_type": oldGenes[gene[2]]["type"][0],
                    "ref_location": oldGenes[gene[2]]["location"],
                    "query_location": None,
                    "query_id": None,
                    "query_type": None,
                    "ref_id": gene[2],
                    "cdsAED": "1.000",
                    "exonAED": "1.000",
                    "ref_transcripts": len(oldGenes[gene[2]]["ids"]),
                    "query_transcripts": 0,
                    "ref_strand": oldGenes[gene[2]]["strand"],
                    "query_strand": None,
                }
    total_cdsAED = []
    total_exonAED = []

    def _sortDict(d):
        return (d[1]["contig"], d[1]["location"][0])

    # sort the annotations by contig and start location
    sGenes = natsorted(iter(result.items()), key=_sortDict)
    sortedGenes = OrderedDict(sGenes)
    out.write(
        "Reference_Location\tReference_ID\tRef_strand\tRef_Num_Transcripts\tQuery_Location\tQuery_ID\tQuery_strand\tQuery_Num_Transcripts\tmRNA_AED\tCDS_AED\n"
    )
    for k, v in list(sortedGenes.items()):
        Rstart = str(v["location"][0])
        Rend = str(v["location"][1])
        if v["query_id"]:
            Qstart = str(v["query_location"][0])
            Qend = str(v["query_location"][1])
        else:
            Qstart = "None"
            Qend = "None"
        total_cdsAED.append(float(v["cdsAED"]))
        total_exonAED.append(float(v["exonAED"]))
        out.write(
            "{:}:{:}-{:}\t{:}\t{:}\t{:}\t{:}:{:}-{:}\t{:}\t{:}\t{:}\t{:}\t{:}\n".format(
                v["contig"],
                Rstart,
                Rend,
                v["ref_id"],
                v["ref_strand"],
                v["ref_transcripts"],
                v["contig"],
                Qstart,
                Qend,
                v["query_id"],
                v["query_strand"],
                v["query_transcripts"],
                v["exonAED"],
                v["cdsAED"],
            )
        )
    Avg_cdsAED = sum(total_cdsAED) / float(len(total_cdsAED))
    Avg_exonAED = sum(total_exonAED) / float(len(total_exonAED))
    totalPident = 0.00
    if totalmatches > 0:
        totalPident = totalmatches / totallength
    if output:
        out.close()
    return [
        NumOldLoci,
        NumOldGenes,
        NumOldmRNA,
        NumOldtRNALoci,
        NumOldtRNA,
        refUnique,
        no_change,
        identicalCDS,
        0.000,
        0.000,
        1,
        NumNewLoci,
        NumNewGenes,
        NumNewmRNA,
        NumNewtRNALoci,
        NumNewtRNA,
        queryUnique,
        no_change,
        identicalCDS,
        Avg_exonAED,
        Avg_cdsAED,
        totalPident,
    ]



import sys
import json
from .utils import check_inputs, zopen
from .gff import gff2dict
from .genbank import tbl2dict


def stats(args):
    check_inputs([args.input] + [args.fasta])
    if not args.input_format:  # we have to guess
        if args.input.endswith((".tbl", ".tbl.gz")):
            args.input_format = "tbl"
        elif args.input.endswith((".gff", ".gff3", ".gff.gz", ".gff3.gz")):
            args.input_format = "gff3"
        else:
            sys.stderr.write(
                "Error: unable to determine -i,--input format: {}".format(args.input)
            )
            sys.exit(1)
    # okay now we can load and convert
    if args.input_format == "tbl":
        Genes, parse_errors = tbl2dict(args.input, args.fasta, {}, table=1)
    elif args.input_format == "gff3":
        Genes = gff2dict(args.input, args.fasta, table=1, debug=args.debug)
    # calculate annotation stats and output
    annot_stats = annotation_stats(Genes)
    if args.out:
        with zopen(args.out, mode="w") as outfile:
            outfile.write(json.dumps(annot_stats, indent=4))
    else:
        sys.stdout.write(json.dumps(annot_stats, indent=4))


def annotation_stats(Genes):
    """
    function to output annotation stats from annotation dictionary
    """
    stats = {
        "genes": 0,
        "common_name": 0,
        "mRNA": 0,
        "tRNA": 0,
        "ncRNA": 0,
        "rRNA": 0,
        "repeat_region": 0,
        "misc_feature": 0,
        "avg_gene_length": 0,
        "transcript-level": {
            "CDS_transcripts": 0,
            "CDS_five_utr": 0,
            "CDS_three_utr": 0,
            "CDS_no_utr": 0,
            "CDS_five_three_utr": 0,
            "CDS_complete": 0,
            "CDS_no-start": 0,
            "CDS_no-stop": 0,
            "CDS_no-start_no-stop": 0,
            "total_exons": 0,
            "total_cds_exons": 0,
            "multiple_exon_transcript": 0,
            "single_exon_transcript": 0,
            "avg_exon_length": 0,
            "avg_protein_length": 0,
            "functional": {
                "go_terms": 0,
                "interproscan": 0,
                "eggnog": 0,
                "pfam": 0,
                "cazyme": 0,
                "merops": 0,
                "busco": 0,
                "secretion": 0,
            },
        },
    }
    # parse annotation dictionary
    if len(Genes) > 0:
        protLengths = []
        geneLengths = []
        exonLengths = []
        for k, v in Genes.items():
            stats["genes"] += 1
            gLength = v["location"][1] - v["location"][0]
            geneLengths.append(gLength)
            if "tRNA" in v["type"]:
                stats["tRNA"] += 1
            elif "rRNA" in v["type"]:
                stats["rRNA"] += 1
            elif "repeat_region" in v["type"]:
                stats["repeat_region"] += 1
            elif v["type"] == "misc_feature":
                stats["misc_feature"] += 1
            elif not "mRNA" in v["type"] and "ncRNA" in v["type"]:
                stats["ncRNA"] += 1
            if v["name"]:
                stats["common_name"] += 1
            for i in range(0, len(v["ids"])):
                if v["type"][i] == "mRNA":
                    stats["mRNA"] += 1
                    stats["transcript-level"]["CDS_transcripts"] += 1
                    pLen = len(v["protein"][i])
                    if v["protein"][i].endswith("*"):
                        pLen -= 1
                    protLengths.append(pLen)
                    try:
                        if len(v["mRNA"][i]) > 1:
                            stats["transcript-level"]["multiple_exon_transcript"] += 1
                        for y in v["mRNA"][i]:
                            exon_length = y[1] - y[0]
                            exonLengths.append(exon_length)
                    except IndexError:
                        sys.stderr.write(
                            "ERROR calculating exon length for {}:\n{}".format(
                                k, json.dumps(v, indent=4)
                            )
                        )
                        sys.exit(1)
                    else:
                        stats["transcript-level"]["single_exon_transcript"] += 1
                    stats["transcript-level"]["total_exons"] += len(v["mRNA"][i])
                    stats["transcript-level"]["total_exons"] += len(v["5UTR"][i])
                    stats["transcript-level"]["total_exons"] += len(v["3UTR"][i])
                    stats["transcript-level"]["total_cds_exons"] += len(v["CDS"][i])
                    if v["partialStart"][i] and v["partialStop"][i]:
                        stats["transcript-level"]["CDS_no-start_no-stop"] += 1
                    elif v["partialStart"][i]:
                        stats["transcript-level"]["CDS_no-start"] += 1
                    elif v["partialStop"][i]:
                        stats["transcript-level"]["CDS_no-stop"] += 1
                    else:
                        stats["transcript-level"]["CDS_complete"] += 1
                    if len(v["5UTR"][i]) > 0 and len(v["3UTR"][i]) > 0:
                        stats["transcript-level"]["CDS_five_three_utr"] += 1
                    elif len(v["3UTR"][i]) > 0:
                        stats["transcript-level"]["CDS_three_utr"] += 1
                    elif len(v["5UTR"][i]) > 0:
                        stats["transcript-level"]["CDS_three_utr"] += 1
                    else:
                        stats["transcript-level"]["CDS_no_utr"] += 1
                    if v["go_terms"][i]:
                        stats["transcript-level"]["functional"]["go_terms"] += 1
                    if any(s.startswith("PFAM:") for s in v["db_xref"][i]):
                        stats["transcript-level"]["functional"]["pfam"] += 1
                    if any(s.startswith("InterPro:") for s in v["db_xref"][i]):
                        stats["transcript-level"]["functional"]["interproscan"] += 1
                    if any(s.startswith("EggNog:") for s in v["note"][i]):
                        stats["transcript-level"]["functional"]["eggnog"] += 1
                    if any(s.startswith("CAZy:") for s in v["note"][i]):
                        stats["transcript-level"]["functional"]["cazyme"] += 1
                    if any(s.startswith("MEROPS:") for s in v["note"][i]):
                        stats["transcript-level"]["functional"]["merops"] += 1
                    if any(s.startswith("BUSCO:") for s in v["note"][i]):
                        stats["transcript-level"]["functional"]["busco"] += 1
                    if any(s.startswith("SECRETED:") for s in v["note"][i]):
                        stats["transcript-level"]["functional"]["secretion"] += 1

        stats["avg_gene_length"] = round(sum(geneLengths) / float(len(geneLengths)), 2)
        stats["transcript-level"]["avg_protein_length"] = round(
            sum(protLengths) / float(len(protLengths)), 2
        )
        try:
            stats["transcript-level"]["avg_exon_length"] = round(
                sum(exonLengths) / float(len(exonLengths)), 2
            )
        except ZeroDivisionError:
            stats["transcript-level"]["avg_exon_length"] = 0
    return stats



import sys
import os
import argparse
import shutil
from .__version__ import __version__
import textwrap as _textwrap


END_FORMATTING = "\033[0m"
BOLD = "\033[1m"
DIM = "\033[2m"
SUPPRESS = "==SUPPRESS=="
OPTIONAL = "?"
ZERO_OR_MORE = "*"
ONE_OR_MORE = "+"
PARSER = "A..."
REMAINDER = "..."
_UNRECOGNIZED_ARGS_ATTR = "_unrecognized_args"


class MyParser(argparse.ArgumentParser):
    """
    This subclass of ArgumentParser changes the error messages, such that if a command is run with
    no other arguments, it will display the help text. If there is a different error, it will give
    the normal response (usage and error).
    """

    def error(self, message):
        if len(sys.argv) == 2:  # if a command was given but nothing else
            self.print_help(file=sys.stderr)
            sys.exit(2)
        else:
            super().error(message)


class MyHelpFormatter(argparse.HelpFormatter):
    """
    This is a custom formatter class for argparse. It allows for some custom formatting,
    in particular for the help texts with multiple options (like bridging mode and verbosity level).
    http://stackoverflow.com/questions/3853722
    """

    def __init__(self, prog):
        terminal_width = shutil.get_terminal_size().columns
        os.environ["COLUMNS"] = str(terminal_width)
        max_help_position = min(max(24, terminal_width // 3), 40)
        super().__init__(prog, max_help_position=max_help_position)

    def _get_help_string(self, action):
        """
        Override this function to add default values, but only when 'default' is not already in the
        help text.
        """
        help_text = action.help
        if action.default != argparse.SUPPRESS and action.default is not None:
            if "default" not in help_text.lower():
                help_text += " (default: {})".format(action.default)
            elif "default: DEFAULT" in help_text:
                help_text = help_text.replace(
                    "default: DEFAULT", "default: {}".format(action.default)
                )
        return help_text

    def _split_lines(self, text, width):
        """
        Override this method to add special behaviour for help texts that start with:
          'R|' - loop text one option per line
        """
        return argparse.HelpFormatter._split_lines(self, text, width)

    def _format_action(self, action):
        result = super(MyHelpFormatter, self)._format_action(action)
        if isinstance(action, argparse._SubParsersAction):
            return "%*s%s" % (self._current_indent, "", result.lstrip())
        return result

    def _format_action_invocation(self, action):
        if isinstance(action, argparse._SubParsersAction):
            return ""
        return super(MyHelpFormatter, self)._format_action_invocation(action)

    def _iter_indented_subactions(self, action):
        if isinstance(action, argparse._SubParsersAction):
            try:
                get_subactions = action._get_subactions
            except AttributeError:
                pass
            else:
                for subaction in get_subactions():
                    yield subaction
        else:
            for subaction in super(MyHelpFormatter, self)._iter_indented_subactions(
                action
            ):
                yield subaction

    def _fill_text(self, text, width, indent):
        text = self._whitespace_matcher.sub(" ", text).strip()
        paragraphs = text.split("|n")
        multiline_text = ""
        for paragraph in paragraphs:
            formatted_paragraph = (
                _textwrap.fill(
                    paragraph, width, initial_indent=indent, subsequent_indent=indent
                )
                + "\n"
            )
            multiline_text = multiline_text + formatted_paragraph
        return multiline_text

    def _format_args(self, action, default_metavar):
        get_metavar = self._metavar_formatter(action, default_metavar)
        if action.nargs is None:
            result = "%s" % get_metavar(1)
        elif action.nargs == OPTIONAL:
            result = "[%s]" % get_metavar(1)
        elif action.nargs == ZERO_OR_MORE:
            metavar = get_metavar(1)
            if len(metavar) == 2:
                result = "[%s [%s ...]]" % metavar
            else:
                result = "[%s ...]" % metavar
        elif action.nargs == ONE_OR_MORE:
            result = "%s" % get_metavar(1)
        elif action.nargs == REMAINDER:
            result = "..."
        elif action.nargs == PARSER:
            result = "%s ..." % get_metavar(1)
        elif action.nargs == SUPPRESS:
            result = ""
        else:
            try:
                formats = ["%s" for _ in range(action.nargs)]
            except TypeError:
                raise ValueError("invalid nargs value") from None
            result = " ".join(formats) % get_metavar(action.nargs)
        return result



import sys
import os
from collections import defaultdict, Counter, OrderedDict
from .interlap import InterLap
from .utils import zopen, check_inputs
from .gff import gff2dict
from .fasta import fastaparser
import numpy as np
import itertools
import json
import random
from natsort import natsorted
from .log import startLogging, system_info


def consensus(args):
    logger = startLogging(logfile=args.logfile)
    global log
    if args.silent:
        log = logger.debug
    else:
        log = logger.info
    system_info(log)
    logger.info(args)
    check_inputs(
        [args.fasta] + args.genes + args.proteins + args.transcripts + args.repeats
    )
    log("GFFtk consensus will generate the best gene model at each locus")
    if args.debug:
        args.debug = args.out + ".all_gene_models.gff3"
        if os.path.isfile(args.debug):
            os.remove(args.debug)
    if args.weights:
        WEIGHTS = {}
        for x in args.weights:
            if ":" in x:
                source, w = x.split(":", 1)
            else:
                source = x
                w = 1
            WEIGHTS[source] = int(w)
    else:
        WEIGHTS = {}
    log("Parsing GFF3 files and clustering data into strand specific loci")
    data = parse_data(args.fasta, args.genes, args.proteins, args.transcripts)
    order, n_evidence = calculate_source_order(data)
    if n_evidence > 0:
        log(
            "Filtered gene models for evidence: {} loci have >3 genes and >2 alignment evidence".format(
                n_evidence
            )
        )
        log(
            "Using these filtered loci, the calculated gene model source weights to use as tiebreakers: {}".format(
                json.dumps(order)
            )
        )
    consensus = {}
    counter = 1
    if args.debug:
        locus_bed = zopen(args.out + ".loci.bed", mode="w")
    else:
        locus_bed = open(os.devnull, "w")
    for contig, obj in data.items():
        for strand in ["+", "-"]:
            for locus in obj[strand]:
                name = "locus_{}".format(counter)
                locus_bed.write(
                    "{}\t{}\t{}\t{}\n".format(
                        contig, locus["locus"][0], locus["locus"][1], name
                    )
                )
                keep = best_model(
                    name,
                    contig,
                    strand,
                    locus,
                    weights=WEIGHTS,
                    order=order,
                    debug=args.debug,
                )
                if not keep:
                    continue
                if not contig in consensus:
                    consensus[contig] = {}
                if not keep[0] in consensus[contig]:
                    consensus[contig][keep[0]] = {
                        "strand": strand,
                        "locus": name,
                        "codon_start": keep[1]["codon_start"],
                        "coords": keep[1]["coords"],
                        "score": keep[1]["score"],
                        "source": keep[1]["source"],
                    }
                counter += 1
    locus_bed.close()
    # check for minimum score
    if not args.minscore:
        score_threshold = auto_score_threshold(WEIGHTS, order)
    else:
        score_threshold = int(args.score_threshold)
    log("Setting minimum gene model score to {}".format(score_threshold))
    sources = {}
    filtered = {}
    total = 0
    for c, o in consensus.items():
        for m, v in o.items():
            if v["score"] > score_threshold:
                total += 1
                if not v["source"] in sources:
                    sources[v["source"]] = 1
                else:
                    sources[v["source"]] += 1
                v["contig"] = c
                v["name"] = m
                v["location"] = (min(min(v["coords"])), max(max(v["coords"])))
                filtered[m] = v

    # now we can filter models in repeat regions
    if args.repeats:
        final, dropped_n = filter_models_repeats(
            args.fasta, args.repeats, filtered, filter_threshold=args.repeat_overlap
        )
        log("{} gene models were dropped due to repeat overlap".format(dropped_n))
    else:
        final = filtered
    log(
        "{} consensus gene models derived from these sources: {}".format(
            len(final), json.dumps(sources)
        )
    )
    # finally we can write to GFF3 format
    gff_writer(final, args.out)
    log("GFFtk consensus is finished: {}".format(args.out))


def filter_models_repeats(fasta, repeats, gene_models, filter_threshold=90):
    dropped = 0
    # return fasta length for stats generation
    seq_length = fasta_length(fasta)
    # build interlap object with repeats
    repeat_length = 0
    repeat_inter = defaultdict(InterLap)
    for r in repeats:
        if os.path.isfile(r):
            if r.endswith(".bed"):
                repeat_inter, repeat_length = bed2interlap(r, inter=repeat_inter)
            else:
                repeat_inter, repeat_length = gff2interlap(r, fasta, inter=repeat_inter)
    pct_repeats = repeat_length / seq_length
    log(
        "Loaded repeats representing {:.2%} of the genome and filtering out loci that are > {}% overlap with repeats".format(
            pct_repeats, filter_threshold
        )
    )
    # now we can filter results by going to through gene models and
    # looking at which models overlap with repeats
    filtered = {}
    for k, v in gene_models.items():
        hits = list(repeat_inter[v["contig"]].find(v["location"]))
        if len(hits) > 0:
            gene_length = v["location"][1] - v["location"][0]
            repeat_overlap = 0
            for x in hits:
                if v["location"][0] > x[0]:
                    s = v["location"][0]
                else:
                    s = x[0]
                if v["location"][1] < x[1]:
                    e = v["location"][1]
                else:
                    e = x[1]
                repeat_overlap += e - s
            gene_repeat_overlap = repeat_overlap / gene_length
            if gene_repeat_overlap <= (filter_threshold / 100):
                filtered[k] = v
            else:
                dropped += 1
        else:
            filtered[k] = v
    return filtered, dropped


def fasta_length(fasta):
    length = 0
    with zopen(fasta) as infile:
        for title, seq in fastaparser(infile):
            length += len(seq)
    return length


def gff2interlap(input, fasta, inter=False):
    """
    function to parse GFF3 file, construct scaffold/gene interlap dictionary
    """
    length = 0
    if not inter:
        inter = defaultdict(InterLap)
    Genes = gff2dict(input, fasta)
    for k, v in natsorted(list(Genes.items())):
        inter[v["contig"]].add((v["location"][0], v["location"][1]))
        length += v["location"][0] - v["location"][1]
    return inter, length


def bed2interlap(bedfile, inter=False):
    # load interlap object from a bed file
    length = 0
    if not inter:
        inter = defaultdict(InterLap)
    with zopen(bedfile) as infile:
        for line in infile:
            line = line.strip()
            chr, start, end = line.split("\t")[:3]
            inter[chr].add((int(start), int(end)))
            length += int(end) - int(start)
    return inter, length


def auto_score_threshold(weights, order, user_weight=6):
    # figure out minimum score
    # min_weights = min(weights.values())
    # min_order = min(order.values())
    allweights = {}
    for w in order.keys():
        allweights[w] = weights.get(w, 1) * user_weight
    min_weights = min(allweights.values())
    threshold = 1 + min_weights
    return threshold


def cluster_interlap(obj):
    seen = set()
    results = []
    for i, loc in enumerate(obj):
        if loc[2] in seen:
            continue
        source = loc[3]
        hits = list(obj.find((loc[0], loc[1])))
        if len(hits) > 1:
            for x in hits:
                seen.add(x[2])
            # sub cluster for unique sources
            cleaned = [hits] #sub_cluster(hits)
            for r in cleaned:
                starts = [x[0] for x in r]
                ends = [x[1] for x in r]
                data = [x[2:] for x in r]
                results.append(
                    {
                        "locus": (min(starts), max(ends)),
                        "genes": data,
                        "proteins": [],
                        "transcripts": [],
                        "repeats": [],
                    }
                )
        else:
            results.append(
                {
                    "locus": (loc[0], loc[1]),
                    "genes": [loc[2:]],
                    "proteins": [],
                    "transcripts": [],
                    "repeats": [],
                }
            )
    return results


def sub_cluster(obj):
    # input will be a list of lists for each locus
    # need to check the source and split into sub clusters if any source > 1
    sources = [x[3] for x in obj]
    s, n = Counter(sources).most_common(1)[0]
    if n == 1:  # then no duplicate sources, return
        return [obj]
    else:  # now we need to split these up
        seeds = []
        queries = []
        for x in obj:
            if x[3] == s:
                seeds.append([x])
            else:
                queries.append(x)
        # now we can assign the queries to seeds
        for y in queries:
            scores = []
            for w in seeds:
                o = get_overlap([y[0], y[1]], [w[0][0], w[0][1]])
                scores.append(o)
            idx = scores.index(max(scores))
            seeds[idx].append(y)
        return seeds


def get_overlap(a, b):
    return max(0, min(a[1], b[1]) - max(a[0], b[0]))


def get_loci(annot_dict):
    # input is annotation dictionary
    plus_inter = defaultdict(InterLap)
    minus_inter = defaultdict(InterLap)
    contigs = []
    pseudo = []
    for gene, info in annot_dict.items():
        if not info["contig"] in contigs:
            contigs.append(info["contig"])
        if info["pseudo"] == True:
            pseudo.append((gene, info))
            continue
        if info["protein"][0].count("*") > 1:
            pseudo.append((gene, info))
            continue
        if info["strand"] == "+":
            plus_inter[info["contig"]].add(
                [
                    info["location"][0],
                    info["location"][1],
                    gene,
                    info["source"],
                    info["CDS"],
                    info["codon_start"][0],
                ]
            )
        else:
            minus_inter[info["contig"]].add(
                [
                    info["location"][0],
                    info["location"][1],
                    gene,
                    info["source"],
                    info["CDS"],
                    info["codon_start"][0],
                ]
            )
    loci = {}
    n_loci = 0
    for chr in contigs:
        plus_bins = {}
        minus_bins = {}
        if chr in plus_inter:
            plus_bins = cluster_interlap(plus_inter[chr])
        n_loci += len(plus_bins)
        if chr in minus_inter:
            minus_bins = cluster_interlap(minus_inter[chr])
        n_loci += len(minus_bins)
        loci[chr] = {"+": plus_bins, "-": minus_bins}
    return loci, n_loci, pseudo


def gffevidence2dict(file, Evi):
    """
    general function to take a evidence in GFF3 format and return locustag/ID dictionary
    """
    seen = set()
    with zopen(file, mode="r") as input:
        for line in input:
            if line.startswith("\n") or line.startswith("#"):
                continue
            line = line.rstrip()
            if line in seen:
                continue
            try:
                (
                    contig,
                    source,
                    feature,
                    start,
                    end,
                    score,
                    strand,
                    phase,
                    attributes,
                ) = line.split("\t")
            except:
                continue
            seen.add(line)
            start = int(start)
            end = int(end)
            info = attributes.split(";")
            fields = {}
            for field in info:
                try:
                    k, v = field.split("=")
                    fields[k] = v
                except (IndexError, ValueError) as E:
                    pass
            if not "ID" in fields:
                continue
            ID = fields["ID"]
            if "target" in fields:
                Target = fields["target"]
            else:
                Target = None
            if not ID in Evi:
                Evi[ID] = {
                    "target": Target,
                    "type": feature,
                    "source": source,
                    "strand": strand,
                    "phase": [phase],
                    "contig": contig,
                    "coords": [(start, end)],
                    "location": (start, end),
                    "score": [score],
                }
            else:
                Evi[ID]["coords"].append((start, end))
                Evi[ID]["score"].append(score)
                Evi[ID]["phase"].append(phase)
                if start < Evi[ID]["location"][0]:
                    Evi[ID]["location"] = (start, Evi[ID]["location"][1])
                if end > Evi[ID]["location"][1]:
                    Evi[ID]["location"] = (Evi[ID]["location"][0], end)
    return Evi


def add_evidence(loci, evidence, source="proteins"):
    # here we will build interlap obj of evidence
    # then loop through loci and pull in evidence that aligns to each locus
    plus_inter = defaultdict(InterLap)
    minus_inter = defaultdict(InterLap)
    for k, v in evidence.items():
        if v["strand"] == "+":
            plus_inter[v["contig"]].add(
                [v["location"][0], v["location"][1], k, v["source"], v["coords"]]
            )
        else:
            minus_inter[v["contig"]].add(
                [v["location"][0], v["location"][1], k, v["source"], v["coords"]]
            )
    for k, v in loci.items():
        for i, x in enumerate(v["+"]):
            hits = list(plus_inter[k].find(x["locus"]))
            if len(hits) > 0:
                # print(k, '+', x['locus'], i, hits)
                cleaned = []
                for h in hits:
                    cleaned.append(h[2:])
                loci[k]["+"][i][source] = cleaned
        for z, y in enumerate(v["-"]):
            hits2 = list(minus_inter[k].find(y["locus"]))
            if len(hits) > 0:
                # print(k, '-', x['locus'], z, hits)
                cleaned = []
                for h in hits2:
                    cleaned.append(h[2:])
                loci[k]["-"][z][source] = cleaned


def parse_data(genome, gene, protein, transcript):
    # parse the input data and build locus data structure
    # all inputs should be lists to support multiple inputs
    Genes = {}
    for g in gene:
        Genes = gff2dict(os.path.abspath(g), genome, annotation=Genes)
    Proteins = {}
    if protein:
        for p in protein:
            Proteins = gffevidence2dict(os.path.abspath(p), Proteins)
    Transcripts = {}
    if transcript:
        for t in transcript:
            Transcripts = gffevidence2dict(os.path.abspath(t), Transcripts)
    # now build loci from gene models
    loci, n_loci, pseudo = get_loci(Genes)
    log(
        "Parsed {} gene models into {} loci. Dropped {} genes models that were pseudo [labled as such or internal stop codons]".format(
            len(Genes), n_loci, len(pseudo)
        )
    )
    # add some evidence
    if len(Proteins) > 0:
        add_evidence(loci, Proteins, source="proteins")
    if len(Transcripts) > 0:
        add_evidence(loci, Transcripts, source="transcripts")
    return loci


def filter4evidence(data, n_genes=3, n_evidence=2):
    # here I want to return loci with at least n_genes and n_evidences
    filt = {}
    n_filt = 0
    for contig, obj in data.items():
        for strand in ["+", "-"]:
            for m in obj[strand]:
                if len(m["genes"]) >= n_genes:
                    if (len(m["transcripts"]) + len(m["proteins"])) >= n_evidence:
                        n_filt += 1
                        if not contig in filt:
                            if strand == "+":
                                filt[contig] = {"+": [m], "-": []}
                            else:
                                filt[contig] = {"-": [m], "+": []}
                        else:
                            if not strand in filt[contig]:
                                filt[contig][strand] = [m]
                            else:
                                filt[contig][strand].append(m)
    return filt, n_filt


def score_by_evidence(locus, weights={}):
    results = {}
    for gene in locus["genes"]:
        score = {"proteins": [], "transcripts": []}
        name, source, coords, cstart = gene
        for s in ["proteins", "transcripts"]:
            for x in locus[s]:
                q_name, q_source, q_coords = x
                score[s].append(
                    score_evidence(coords[0], q_coords, weight=weights.get(source, 1))
                )
        results[name] = {
            "protein_evidence_score": sum(score["proteins"]),
            "transcript_evidence_score": sum(score["transcripts"]),
        }
    return results


def order_sources(locus):
    results = {}
    # print(locus['genes'])
    # first lets check if all predictions are the same
    allcoords = [x[2][0] for x in locus["genes"]]
    if len([list(i) for i in set(tuple(i) for i in allcoords)]) > 0:
        for gene in locus["genes"]:
            score = {"proteins": [], "transcripts": []}
            name, source, coords, cstart = gene
            for s in ["proteins", "transcripts"]:
                for x in locus[s]:
                    q_name, q_source, q_coords = x
                    score[s].append(score_evidence(coords[0], q_coords, weight=1))
            results[name] = {
                "source": source,
                "coords": coords[0],
                "score": sum(score["proteins"]) + sum(score["transcripts"]),
            }
    else:
        results = {
            locus["genes"][0][0]: {
                "source": locus["genes"][0][1],
                "coords": locus["genes"][0][2],
                "score": 0,
            }
        }
    return results


def debug_gff_writer(outfile, contig, strand, data):
    # input is a sorted list of gene models from best_model function
    with zopen(outfile, mode="a") as out:
        for x in data:
            name, d = x
            if strand == "+":
                sortedCoords = sorted(d["coords"], key=lambda tup: tup[0])
            else:
                sortedCoords = sorted(d["coords"], key=lambda tup: tup[0], reverse=True)
            g_start = min(min(sortedCoords))
            g_end = max(max(sortedCoords))
            out.write(
                "{}\t{}\tgene\t{}\t{}\t{:.4f}\t{}\t.\tID={};Note={},{},reasonable:{};\n".format(
                    contig,
                    d["source"],
                    g_start,
                    g_end,
                    d["score"],
                    strand,
                    name,
                    d["locus"],
                    d["all_scores"],
                    d["check"],
                )
            )
            out.write(
                "{}\t{}\tmRNA\t{}\t{}\t.\t{}\t.\tID={}.mrna;Parent={};Note={},{},reasonable:{};\n".format(
                    contig,
                    d["source"],
                    g_start,
                    g_end,
                    strand,
                    name,
                    name,
                    d["locus"],
                    d["all_scores"],
                    d["check"],
                )
            )
            # now write exon and CDS features
            num_exons = len(sortedCoords)
            for x in range(0, num_exons):
                ex_num = x + 1
                out.write(
                    "{:}\t{:}\texon\t{:}\t{:}\t.\t{:}\t.\tID={:}.exon{:};Parent={:}.mrna;\n".format(
                        contig,
                        d["source"],
                        sortedCoords[x][0],
                        sortedCoords[x][1],
                        strand,
                        name,
                        ex_num,
                        name,
                    )
                )
            current_phase = d["codon_start"] - 1
            for y in range(0, num_exons):
                out.write(
                    "{:}\t{:}\tCDS\t{:}\t{:}\t.\t{:}\t{:}\tID={:}.cds;Parent={:}.mrna;\n".format(
                        contig,
                        d["source"],
                        sortedCoords[y][0],
                        sortedCoords[y][1],
                        strand,
                        current_phase,
                        name,
                        name,
                    )
                )
                current_phase = (
                    current_phase
                    - (int(sortedCoords[y][1]) - int(sortedCoords[y][0]) + 1)
                ) % 3
                if current_phase == 3:
                    current_phase = 0


def reasonable_model(
    coords, min_protein=30, min_exon=3, min_intron=10, max_intron=-1, max_exon=-1
):
    # look at coordinates and check some stats
    exon_lengths = []
    intron_lengths = []
    sortedCoords = sorted(coords, key=lambda tup: tup[0])
    for i, x in enumerate(sortedCoords):
        exon_lengths.append(x[1] - x[0])
        if i > 0:
            intron_lengths.append(x[0] - sortedCoords[i - 1][1])
    if min(exon_lengths) < min_exon:
        return "exon < {}".format(min_exon)
    if max_exon > 0 and max(exon_lengths) > max_exon:
        return "exon > {}".format(max_exon)
    if len(intron_lengths) > 0:
        if min(intron_lengths) < min_intron:
            return "intron < {}".format(min_intron)
        if max_intron > 0 and max(intron_lengths) > max_intron:
            return "intron > {}".format(max_intron)
    if (sum(exon_lengths) / 3) < min_protein:
        return "protein is < {} amino acids".format(min_protein)
    return True


def best_model(
    locus_name,
    contig,
    strand,
    locus,
    debug=False,
    weights={},
    order={},
    min_exon=3,
    min_intron=10,
    max_intron=-1,
    max_exon=-1,
):
    results = {}
    # check if could be multiple
    sources = [x[1] for x in locus["genes"]]
    s, n = Counter(sources).most_common(1)[0]
    if n > 1:  # could be multiple genes here
        print('---------------')
        print(s, n)
        print('Multiple sources one locus: {}'.format(locus["genes"]))
    # get evidence scores
    evidence_scores = score_by_evidence(locus)
    # get aed scores
    de_novo_aed_scores = de_novo_distance(locus)
    # now go through genes and combine scores
    for gene in locus["genes"]:
        name, source, coords, cstart = gene
        # check if gene is reasonable
        check = reasonable_model(
            coords[0],
            min_exon=min_exon,
            min_intron=min_intron,
            max_intron=max_intron,
            max_exon=max_exon,
        )
        user_score = weights.get(source, 1) * 6
        order_score = order.get(source, 0)
        aed_score = de_novo_aed_scores.get(name, 0) * 3
        evi = evidence_scores.get(name)
        prot_score = evi["protein_evidence_score"]
        trans_score = evi["transcript_evidence_score"]
        # if user_score is 0, then make total score 0 as user has indicated not to use those models
        if user_score == 0:
            total_score = 0
        else:
            total_score = user_score + aed_score + prot_score + trans_score
        results[name] = {
            "source": source,
            "coords": coords[0],
            "check": check,
            "score": total_score,
            "locus": locus_name,
            "codon_start": cstart,
            "all_scores": "user:{},aed:{:.4f},prot:{},trans:{},order:{}".format(
                user_score, aed_score, prot_score, trans_score, order_score
            ),
        }
    best_result = sorted(results.items(), key=lambda e: e[1]["score"], reverse=True)
    # if debug than write GFF3 to file
    if debug:
        debug_gff_writer(debug, contig, strand, best_result)
    best_result_filtered = [x for x in best_result if x[1]["check"] == True]
    if n > 1:
        print(best_result_filtered)
    # check if we need to break any ties
    if len(best_result_filtered) > 0:
        best_score = best_result_filtered[0][1]["score"]
        anyties = [x for x in best_result_filtered if x[1]["score"] >= best_score]
        if len(anyties) > 1:
            breaktie = sorted(
                anyties, key=lambda x: order[x[1]["source"]], reverse=True
            )
            bm = random.choice(breaktie)
        else:
            bm = anyties[0]
    else:
        bm = []
    return bm


def cluster_by_aed(locus, score=0.005):
    distances = calculate_gene_distance(locus)
    # try to cluster, dict is ordered so run through in order
    clusters = {}  # start with identical models
    seen = {}
    for k, v in distances.items():
        # print(k, v)
        if k in seen:
            centroid = seen.get(k)
        else:
            centroid = k
        for g, a in v.items():
            if g in seen:
                centroid = seen.get(g)
            if a <= score:
                if centroid == g:
                    continue
                if not centroid in clusters:
                    clusters[centroid] = [g]
                else:
                    if not g in clusters[centroid]:
                        clusters[centroid].append(g)
                seen[g] = centroid
                if g in clusters:
                    del clusters[g]
            else:
                if not g in clusters and not g in seen:
                    clusters[g] = []
    # print(clusters)
    final = []
    accounted = 0
    for k, v in sorted(clusters.items(), key=lambda x: len(x[1]), reverse=True):
        all = [k] + v
        accounted += len(all)
        final.append(all)
    assert len(locus["genes"]) == accounted
    return final


def map_coords(g_coords, e_coords):
    # map the e_coords onto the g_coords
    r = [
        [],
    ] * len(g_coords)
    refInterlap = InterLap(g_coords)
    for e in e_coords:
        if e in refInterlap:
            hit = list(refInterlap.find(e))
            if len(hit) > 1:
                continue
            i = g_coords.index(hit[0])
            diff = np.subtract(e, hit[0])
            r[i] = list(diff)
    return r


def score_evidence(g_coords, e_coords, weight=2):
    # simple score: 10 to 0
    # 1 == e_coords contained and match intron/exons
    # 0.5 == e_coords partially contained
    # 0 == e_coords not contained or contained but intron/exon boundaries do not match
    # coords are list of tuples
    # g_coords is single gene model
    # e_coords is single evidence coordinates
    if g_coords == e_coords:
        return 10 * weight
    emap = map_coords(g_coords, e_coords)
    score = 0
    if len(g_coords) > 1:
        mult = []
        for x in emap:
            if not x:
                mult.append(0)
            else:
                if x[0] >= 0 and x[1] <= 0:
                    mult.append(10)
                else:
                    mult.append(5)
        score = (sum(mult) // len(g_coords)) * weight
    else:  # single exon gene
        if emap[0]:
            if emap[0][0] >= 0 and emap[0][1] <= 0:
                score = 10 * weight
            else:
                score = 5 * weight
        else:
            score = 0
    return score


def calculate_source_order(data):
    # filter data for evidence and calculate the global accuracy of the sources
    # this will output a rank order of sources to use when there is no evidence
    f_data, n_filt = filter4evidence(data)
    if n_filt > 0:
        cweights = {}
        for contig, obj in f_data.items():
            for strand in ["+", "-"]:
                for locus in obj[strand]:
                    all = order_sources(locus)
                    for k, v in all.items():
                        if not v["source"] in cweights:
                            cweights[v["source"]] = [v["score"]]
                        else:
                            cweights[v["source"]].append(v["score"])
        sweights = []
        avgweights = []
        for k, v in cweights.items():
            aweight = sum(v) // len(v)
            sweights.append((k, aweight))
            avgweights.append(aweight)
        min_weight = min(avgweights)
        order = (
            OrderedDict()
        )  # y[0] for y in sorted(sweights.items(), key=lambda x: x[1], reverse=True)]
        for z in sorted(sweights, key=lambda x: x[1], reverse=True):
            order[z[0]] = z[1]  # round(z[1] / min_weight)
    else:  # there are no loci with evidence, so return all the sources instead and set all to 1
        order = OrderedDict()
        for contig, obj in data.items():
            for strand in ["+", "-"]:
                for locus in obj[strand]:
                    for gene in locus["genes"]:
                        name, source, coords, cstart = gene
                        if not source in order:
                            order[source] = 1
    return order, n_filt


def calculate_gene_distance(locus):
    # caclulate AED for all pairs of gene predictions
    dist = {}
    for y in locus["genes"]:
        dist[y[0]] = {}
    for x in itertools.combinations(locus["genes"], 2):
        a, b = x
        a_name, a_source, a_coord, _ = a
        b_name, b_source, b_coord, _ = b
        aed = getAED(b_coord[0], a_coord[0])
        dist[a_name][b_name] = aed
        dist[b_name][a_name] = aed
    return dist


def de_novo_distance(locus):
    results = {}
    if len(locus["genes"]) > 1:
        # calculate a score for each gene based on AED similaritiy with other models
        distances = calculate_gene_distance(locus)
        # add up the aeds, subtract total aed from total gene models
        # idea here is that higher value is a model that has similarity with other predictions
        # this is not as fine-grained as EVM's exon and splice site overlap method, where each exon/intron is evaluated
        for k, v in distances.items():
            total = float(sum(v.values()))
            # print(locus['genes'], k, v, len(locus['genes']), len(v))
            assert (
                len(locus["genes"]) == len(v) + 1
            ), "ERROR, AED distance calculation failed"
            results[k] = (len(locus["genes"])) - total
    else:  # single gene, so value here is 0
        geneID = locus["genes"][0][0]
        results[geneID] = 0.0
    return results


def getAED(query, reference):
    """
    function to calcuate annotation edit distance between two transcript coordinates
    AED = 1 - (SN + SP / 2)
    SN = fraction of ref predicted
    SP = fraction prediction overlapping the ref
    """

    def _length(listTup):
        len = 0
        for i in listTup:
            l = abs(i[0] - i[1])
            len += l
        return len

    # check if identical
    if query == reference:
        return 0.000
    # make sure sorted
    rLen = _length(reference)
    refInterlap = InterLap(reference)
    QueryOverlap = 0
    qLen = 0
    for exon in query:
        qLen += abs(exon[0] - exon[1])
        if exon in refInterlap:  # exon overlaps at least partially with reference
            hit = list(refInterlap.find(exon))
            for h in hit:
                # will return array of exon minus hit at each pos
                diff = np.subtract(exon, h)
                if diff[0] <= 0 and diff[1] >= 0:  # then query exon covers ref exon
                    cov = abs(h[0] - h[1])
                    QueryOverlap += cov
                elif diff[0] <= 0 and diff[1] < 0:  # means query partial covers ref
                    cov = abs(h[0] - exon[1])
                    QueryOverlap += cov
                elif diff[0] > 0 and diff[1] >= 0:  # means query partial covers ref
                    cov = abs(exon[0] - h[1])
                    QueryOverlap += cov
                elif diff[0] > 0 and diff[1] < 1:
                    cov = abs(exon[0] - exon[1])
                    QueryOverlap += cov
    # calculate AED
    SP = QueryOverlap / float(qLen)
    SN = QueryOverlap / float(rLen)
    AED = 1 - ((SN + SP) / 2)
    return AED


def gff_writer(input, output):
    def _sortDict(d):
        return (d[1]["contig"], d[1]["location"][0])

    # open output for writing and write header
    with zopen(output, mode="w") as outfile:
        outfile.write("##gff-version 3\n")
        # sort the annotations by contig and start location
        sGenes = natsorted(iter(input.items()), key=_sortDict)
        sortedGenes = OrderedDict(sGenes)
        counter = 1
        for k, d in list(sortedGenes.items()):
            locusTag = "CGM_{}".format(str(counter).zfill(6))
            if d["strand"] == "+":
                sortedCoords = sorted(d["coords"], key=lambda tup: tup[0])
            else:
                sortedCoords = sorted(d["coords"], key=lambda tup: tup[0], reverse=True)
            g_start, g_end = d["location"]
            outfile.write(
                "{}\tGFFtk\tgene\t{}\t{}\t{:.4f}\t{}\t.\tID={};Note={} derived from {} model {};\n".format(
                    d["contig"],
                    g_start,
                    g_end,
                    d["score"],
                    d["strand"],
                    locusTag,
                    d["locus"],
                    d["source"],
                    k,
                )
            )
            outfile.write(
                "{}\tGFFtk\tmRNA\t{}\t{}\t.\t{}\t.\tID={}.mrna;Parent={};\n".format(
                    d["contig"], g_start, g_end, d["strand"], locusTag, locusTag
                )
            )
            # now write exon and CDS features
            num_exons = len(sortedCoords)
            current_phase = d["codon_start"] - 1
            for x in range(0, num_exons):
                ex_num = x + 1
                outfile.write(
                    "{}\tGFFtk\texon\t{}\t{}\t.\t{}\t.\tID={}.exon{};Parent={}.mrna;\n".format(
                        d["contig"],
                        sortedCoords[x][0],
                        sortedCoords[x][1],
                        d["strand"],
                        locusTag,
                        ex_num,
                        locusTag,
                    )
                )
                outfile.write(
                    "{}\tGFFtk\tCDS\t{}\t{}\t.\t{}\t{}\tID={}.cds;Parent={}.mrna;\n".format(
                        d["contig"],
                        sortedCoords[x][0],
                        sortedCoords[x][1],
                        d["strand"],
                        current_phase,
                        locusTag,
                        locusTag,
                    )
                )
                current_phase = (
                    current_phase
                    - (int(sortedCoords[x][1]) - int(sortedCoords[x][0]) + 1)
                ) % 3
                if current_phase == 3:
                    current_phase = 0
            counter += 1



import sys
import numpy as np
from natsort import natsorted
from .fasta import getSeqRegions, fasta2dict, translate
from .utils import zopen, readBlocks2
from .go import go_term_dict
from .interlap import InterLap


def tbl2dict(input, fasta, annotation=False, table=1, debug=False):
    """
    need a method to convert directly from NCBI tbl format to several output formats
    to avoid conversion problems with GBK files that have mutliple transcripts
    if can load funannotate dictionary directly from tbl format, then can write the other
    formats directly
    """
    if not annotation:
        annotation = {}
    errors = []
    with zopen(input) as infile:
        contig = ""
        for item in readBlocks2(infile, ">Feature", "\tgene\n"):
            if item[0].startswith(">Feature"):  # this will be contig header block
                contig = item[0].rstrip().split(" ")[-1]
            else:  # these are all gene model blocks
                (
                    geneID,
                    Name,
                    type,
                    start,
                    end,
                    fivepartial,
                    threepartial,
                    strand,
                    location,
                ) = (None,) * 9
                codon_start = []
                transcriptID = []
                proteinID = []
                synonyms = []
                product = []
                phase = "?"
                first, firstpartial, second, secondpartial = (False,) * 4
                position = None
                # check number of transcripts
                tNum = 0
                for z in item:
                    if z.startswith("\t\t\ttranscript_id"):
                        tNum += 1
                if tNum > 0:
                    tNum = int(tNum / 2)
                if tNum == 0:
                    tNum = 1
                # setup lists for transcripts
                mRNA = [[] for y in range(tNum)]
                CDS = [[] for y in range(tNum)]
                note = [[] for y in range(tNum)]
                dbxref = [[] for y in range(tNum)]
                ECnum = [[] for y in range(tNum)]
                go_terms = [[] for y in range(tNum)]
                fivepartial = [
                    False,
                ] * tNum
                threepartial = [
                    False,
                ] * tNum
                currentNum = 0
                for x in item:
                    exonF, exonR, cdsF, cdsR, cols = (None,) * 5
                    if x.endswith("\tgene\n") and not position:
                        cols = x.strip().split("\t")
                        position = "gene"
                        if cols[0].startswith("<"):
                            first = int(cols[0].split("<")[-1])
                        else:
                            first = int(cols[0])
                        if cols[1].startswith(">"):
                            second = int(cols[1].split(">")[-1])
                        else:
                            second = int(cols[1])
                        if first < second:
                            start = first
                            end = second
                            strand = "+"
                        else:
                            start = second
                            end = first
                            strand = "-"
                        location = (start, end)
                    elif x.startswith("\t\t\tgene\t"):
                        Name = x.strip().split("\t")[-1]
                    elif x.startswith("\t\t\tlocus_tag\t"):
                        geneID = x.strip().split("\t")[-1]
                    elif (
                        x.endswith("\ttRNA\n")
                        and x.count("\t") == 2
                        and position == "gene"
                    ):
                        type = "tRNA"
                        position = "tRNA"
                        cols = x.strip().split("\t")
                        exonF = int(cols[0].replace("<", ""))
                        exonR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            mRNA[currentNum].append((exonF, exonR))
                        else:
                            mRNA[currentNum].append((exonR, exonF))
                    elif (
                        x.endswith("\tncRNA\n")
                        and x.count("\t") == 2
                        and position == "gene"
                    ):
                        type = "ncRNA"
                        position = "ncRNA"
                        cols = x.strip().split("\t")
                        exonF = int(cols[0].replace("<", ""))
                        exonR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            mRNA[currentNum].append((exonF, exonR))
                        else:
                            mRNA[currentNum].append((exonR, exonF))
                    elif (
                        x.endswith("\trRNA\n")
                        and x.count("\t") == 2
                        and position == "gene"
                    ):
                        type = "rRNA"
                        position = "rRNA"
                        cols = x.strip().split("\t")
                        exonF = int(cols[0].replace("<", ""))
                        exonR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            mRNA[currentNum].append((exonF, exonR))
                        else:
                            mRNA[currentNum].append((exonR, exonF))
                    elif x.endswith("\tmRNA\n") and x.count("\t") == 2:
                        if position == "CDS":
                            currentNum += 1
                        elif position == "gene":
                            type = "mRNA"
                        position = "mRNA"
                        cols = x.strip().split("\t")
                        exonF = int(cols[0].replace("<", ""))
                        exonR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            mRNA[currentNum].append((exonF, exonR))
                        else:
                            mRNA[currentNum].append((exonR, exonF))
                    elif x.endswith("\tCDS\n") and x.count("\t") == 2:
                        position = "CDS"
                        cols = x.strip().split("\t")
                        cdsF = int(cols[0].replace("<", ""))
                        cdsR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            CDS[currentNum].append((cdsF, cdsR))
                        else:
                            CDS[currentNum].append((cdsR, cdsF))
                    elif x.startswith("\t\t\tcodon_start\t"):
                        cNum = int(x.strip().split("\t")[-1])
                        codon_start.append(cNum)
                        phase = cNum - 1
                    elif x.startswith("\t\t\tproduct\t") and position != "mRNA":
                        product.append(x.strip().split("\t")[-1])
                    elif x.startswith("\t\t\ttranscript_id\t"):
                        tID = x.strip().split("|")[-1]
                        if "_mrna" in tID:
                            tID = tID.replace("_mrna", "")
                        if not tID in transcriptID:
                            transcriptID.append(tID)
                    elif x.startswith("\t\t\tprotein_id\t"):
                        pID = x.strip().split("|")[-1]
                        if not pID in proteinID:
                            proteinID.append(pID)
                    elif x.startswith("\t\t\tgene_synonym\t"):
                        synonyms.append(x.strip().split("\t")[-1])
                    elif x.startswith("\t\t\tgo_"):  # go terms
                        go_terms[currentNum].append(
                            "GO:{:}".format(x.strip().split("|")[1])
                        )
                    elif x.startswith("\t\t\tnote\t"):
                        note[currentNum].append(x.strip().split("\t")[-1])
                    elif x.startswith("\t\t\tdb_xref\t"):
                        dbxref[currentNum].append(x.strip().split("\t")[-1])
                    elif x.startswith("\t\t\tEC_number\t"):
                        ECnum[currentNum].append(x.strip().split("\t")[-1])
                    elif position == "mRNA" and x.count("\t") == 1:
                        cols = x.strip().split("\t")
                        exonF = int(cols[0].replace("<", ""))
                        exonR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            mRNA[currentNum].append((exonF, exonR))
                        else:
                            mRNA[currentNum].append((exonR, exonF))
                    elif position in ["tRNA", "ncRNA", "rRNA"] and x.count("\t") == 1:
                        cols = x.strip().split("\t")
                        exonF = int(cols[0].replace("<", ""))
                        exonR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            mRNA[currentNum].append((exonF, exonR))
                        else:
                            mRNA[currentNum].append((exonR, exonF))
                    elif position == "CDS" and x.count("\t") == 1:
                        cols = x.strip().split("\t")
                        cdsF = int(cols[0].replace("<", ""))
                        cdsR = int(cols[1].replace(">", ""))
                        if strand == "+":
                            CDS[currentNum].append((cdsF, cdsR))
                        else:
                            CDS[currentNum].append((cdsR, cdsF))
                if not geneID in annotation:
                    if type in ["tRNA", "ncRNA", "rRNA"]:
                        annotation[geneID] = {
                            "name": Name,
                            "type": [type,]*tNum,
                            "transcript": [],
                            "cds_transcript": [],
                            "protein": [],
                            "5UTR": [[]],
                            "3UTR": [[]],
                            "codon_start": codon_start,
                            "ids": [geneID + "-T1"],
                            "CDS": CDS,
                            "mRNA": mRNA,
                            "strand": strand,
                            "gene_synonym": synonyms,
                            "location": location,
                            "contig": contig,
                            "product": product,
                            "source": "GFFtk",
                            "phase": [phase],
                            "db_xref": dbxref,
                            "go_terms": go_terms,
                            "EC_number": ECnum,
                            "note": note,
                            "partialStart": [True],
                            "partialStop": [True],
                            "pseudo": False,
                        }
                    else:
                        annotation[geneID] = {
                            "name": Name,
                            "type": [type,]*tNum,
                            "transcript": [],
                            "cds_transcript": [],
                            "protein": [],
                            "5UTR": [],
                            "3UTR": [],
                            "codon_start": codon_start,
                            "ids": proteinID,
                            "CDS": CDS,
                            "mRNA": mRNA,
                            "strand": strand,
                            "gene_synonym": synonyms,
                            "location": location,
                            "contig": contig,
                            "product": product,
                            "source": "GFFtk",
                            "phase": [phase],
                            "db_xref": dbxref,
                            "go_terms": go_terms,
                            "EC_number": ECnum,
                            "note": note,
                            "partialStart": fivepartial,
                            "partialStop": threepartial,
                            "pseudo": False,
                        }
    # now we need to sort coordinates, get protein/transcript sequences and capture UTRs
    SeqRecords = fasta2dict(fasta)
    for k, v in list(annotation.items()):
        # @nextgenusfs we should clarify or rename this variable to indicate
        # i is the i-th transcript, right??
        for i in range(0, len(v["ids"])):
            try:
                featuretype = v["type"][i]
            except IndexError:
                featuretype =  v["type"][0]
            if featuretype in ["mRNA", "tRNA", "ncRNA", "rRNA"]:
                if v["strand"] == "+":
                    sortedExons = sorted(v["mRNA"][i], key=lambda tup: tup[0])
                else:
                    sortedExons = sorted(
                        v["mRNA"][i], key=lambda tup: tup[0], reverse=True
                    )
                annotation[k]["mRNA"][i] = sortedExons
                mrnaSeq = getSeqRegions(SeqRecords, v["contig"], sortedExons)
                annotation[k]["transcript"].append(mrnaSeq)
            if featuretype == "mRNA":
                if v["strand"] == "+":
                    sortedCDS = sorted(v["CDS"][i], key=lambda tup: tup[0])
                else:
                    sortedCDS = sorted(
                        v["CDS"][i], key=lambda tup: tup[0], reverse=True
                    )
                cdsSeq = getSeqRegions(SeqRecords, v["contig"], sortedCDS)
                # If the translation starts in middle of a codon,
                # we need to truncate the CDS seq either at start or end
                # depending on strand.
                if v["codon_start"][i] > 1:
                    if v["strand"] == "+":
                        # drop first N bases based on codon_start
                        # to reflect the translation frame
                        cdsSeq = cdsSeq[v["codon_start"][i] - 1 :]
                    elif v["strand"] == "-":
                        # drop last N bases based on codon_start
                        # to reflect the translation frame (because this is
                        # is reverse strand gene)
                        endTrunc = len(cdsSeq) - (v["codon_start"][i] - 1)
                        cdsSeq = cdsSeq[0:endTrunc]
                    else:
                        # could trigger more of a warning/error
                        errors.append(
                            "ERROR strand (%s) is nonsensical for %s" % (v["strand"], k)
                        )
                annotation[k]["cds_transcript"].append(cdsSeq)
                annotation[k]["CDS"][i] = sortedCDS
                protSeq = translate(cdsSeq, v["strand"], 0, table=table)
                if protSeq:
                    annotation[k]["protein"].append(protSeq)
                    if protSeq.endswith("*"):
                        annotation[k]["partialStop"][i] = False
                    else:
                        annotation[k]["partialStop"][i] = True
                    if v["codon_start"][i] == 1 and protSeq.startswith("M"):
                        annotation[k]["partialStart"][i] = False
                    else:
                        annotation[k]["partialStart"][i] = True
                # get UTRs
                try:
                    FiveUTR, ThreeUTR = findUTRs(sortedCDS, sortedExons, v["strand"])
                    annotation[k]["5UTR"].append(FiveUTR)
                    annotation[k]["3UTR"].append(ThreeUTR)
                except ValueError:
                    errors.append("ERROR", k, v)
    if debug:
        if len(errors) > 0:
            for e in errors:
                sys.stderr.write("{}\n".format(e))
    return annotation, errors


def findUTRs(cds, mrna, strand):
    FiveUTR = []
    ThreeUTR = []
    if cds != mrna:
        inter = InterLap()
        inter.add(cds)
        for i, x in enumerate(mrna):
            if not x in inter:
                loc = (list(inter)[0][0], list(inter)[-1][1])
                diff = np.subtract(x, loc)
                if diff[0] < 0 and diff[1] < 0:
                    if strand == "+":
                        FiveUTR.append(x)
                    else:
                        ThreeUTR.append(x)
                elif diff[0] > 0 and diff[1] > 0:
                    if strand == "+":
                        ThreeUTR.append(x)
                    else:
                        FiveUTR.append(x)
            else:
                hit = list(inter.find(x))
                if x == hit[0]:
                    continue
                else:
                    diff = np.subtract(x, hit[0])
                    if strand == "+":
                        if int(diff[0]) < 1 and int(diff[1]) == 0:
                            FiveUTR.append((x[0], hit[0][0] - 1))
                        elif int(diff[1]) > 1 and int(diff[0]) == 0:
                            ThreeUTR.append((hit[0][1] + 1, x[1]))
                        elif int(diff[0]) < 1 and int(diff[1]) > 1:
                            FiveUTR.append((x[0], hit[0][0] - 1))
                            ThreeUTR.append((hit[0][1] + 1, x[1]))
                    else:
                        if diff[0] == 0 and diff[1] > 0:
                            FiveUTR.append((hit[0][1] + 1, x[1]))
                        elif diff[0] < 0 and diff[1] == 0:
                            ThreeUTR.append((x[0], hit[0][0] - 1))
                        elif diff[0] < 0 and diff[1] > 0:
                            FiveUTR.append((hit[0][1] + 1, x[1]))
                            ThreeUTR.append((x[0], hit[0][0] - 1))
    return FiveUTR, ThreeUTR


def dict2tbl(
    genesDict,
    scaff2genes,
    scaffLen,
    SeqCenter,
    SeqRefNum,
    skipList,
    output=False,
    annotations=False,
    external=False,
):
    """
    function to take funannotate annotation dictionaries and convert to NCBI tbl output
    """
    duplicates = 0
    pseudo = 0
    nocds = 0
    errors = []
    # to parse annotations, will need to have access to GO OBO dictionary
    goDict = {}
    if annotations:
        goDict, go_format, go_date = go_term_dict()

    def _goFormat(id, goDict=goDict):
        # go_function    serine-type endopeptidase activity|0004252||IEA
        # go_process proteolysis|0006508||IEA
        # go_component   nucleus|0005634||IEA
        if id in goDict:
            if goDict[id]["namespace"] == "biological_process":
                base = "go_process"
            elif goDict[id]["namespace"] == "molecular_function":
                base = "go_function"
            elif goDict[id]["namespace"] == "cellular_component":
                base = "go_component"
            reformatted = "\t\t\t{:}\t{:}|{:}||IEA".format(
                base, goDict[id]["name"], id.replace("GO:", "")
            )
            return reformatted
        else:
            return False

    if output:
        tbl = zopen(output, mode="w")
    else:
        tbl = sys.stdout
    # now convert to tbl format
    for k, v in natsorted(list(scaff2genes.items())):
        tbl.write(">Feature %s\n" % k)
        tbl.write("1\t%s\tREFERENCE\n" % scaffLen.get(k))
        tbl.write("\t\t\t%s\t%s\n" % (SeqCenter, SeqRefNum))
        for genes in v:  # now loop through each gene on the scaffold
            if genes in skipList:
                continue
            # single funannotate standard dictionary
            geneInfo = genesDict.get(genes)
            if not geneInfo["ids"]:
                continue
            if (
                not len(geneInfo["ids"])
                == len(geneInfo["mRNA"])
                == len(geneInfo["CDS"])
            ):
                continue
            #print(genes, geneInfo['note'])
            # check for partial models
            if True in geneInfo["partialStart"]:
                ps = "<"
            else:
                ps = ""
            if True in geneInfo["partialStop"]:
                pss = ">"
            else:
                pss = ""
            # now write gene model
            if geneInfo["strand"] == "+":
                tbl.write(
                    "%s%i\t%s%i\tgene\n"
                    % (ps, geneInfo["location"][0], pss, geneInfo["location"][1])
                )
                if annotations:
                    if geneInfo["name"]:
                        tbl.write("\t\t\tgene\t%s\n" % geneInfo["name"])
                    if geneInfo["gene_synonym"]:
                        for alias in geneInfo["gene_synonym"]:
                            tbl.write("\t\t\tgene_synonym\t%s\n" % alias)
                tbl.write("\t\t\tlocus_tag\t%s\n" % genes)
            else:
                tbl.write(
                    "%s%i\t%s%i\tgene\n"
                    % (ps, geneInfo["location"][1], pss, geneInfo["location"][0])
                )
                if annotations:
                    if geneInfo["name"]:
                        tbl.write("\t\t\tgene\t%s\n" % geneInfo["name"])
                    if geneInfo["gene_synonym"]:
                        for alias in geneInfo["gene_synonym"]:
                            tbl.write("\t\t\tgene_synonym\t%s\n" % alias)
                tbl.write("\t\t\tlocus_tag\t%s\n" % genes)

            # now will output the gene models with -T1, -T2, -T3 annotations based on expression values
            # means need to get the order
            order = []
            # multiple transcripts, so get order of highest TPM
            if len(geneInfo["ids"]) > 1:
                tpms = []
                for num, tpm in enumerate(geneInfo["note"]):
                    for item in tpm:
                        if item.startswith("TPM:"):
                            value = float(item.split(":")[-1])
                            tpms.append((value, num))
                if len(tpms) > 0:
                    for x in sorted(tpms, reverse=True):
                        order.append(x[1])
                else:
                    order = list(range(0, len(geneInfo["ids"])))
            else:
                order.append(0)
            for num, i in enumerate(order):  # now write mRNA and CDS features
                if external:
                    protein_id = geneInfo["ids"][i]
                else:
                    protein_id = genes + "-T" + str(num + 1)
                if geneInfo["type"][i] in ["mRNA", "ncRNA"]:
                    ps = "<"
                    pss = ">"
                    if geneInfo["partialStart"][i] is False:
                        ps = ""
                    if geneInfo["partialStop"][i] is False:
                        pss = ""
                    if geneInfo["strand"] == "+":
                        for num, exon in enumerate(geneInfo["mRNA"][i]):
                            # single exon, so slightly differnt method
                            if num == 0 and num == len(geneInfo["mRNA"][i]) - 1:
                                tbl.write(
                                    "%s%s\t%s%s\t%s\n"
                                    % (ps, exon[0], pss, exon[1], geneInfo["type"][i])
                                )
                            elif num == 0:
                                tbl.write(
                                    "%s%s\t%s\t%s\n"
                                    % (ps, exon[0], exon[1], geneInfo["type"][i])
                                )
                            # this is last one
                            elif num == len(geneInfo["mRNA"][i]) - 1:
                                tbl.write("%s\t%s%s\n" % (exon[0], pss, exon[1]))
                            else:
                                tbl.write("%s\t%s\n" % (exon[0], exon[1]))
                        tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
                        tbl.write(
                            "\t\t\ttranscript_id\tgnl|ncbi|%s_mrna\n" % (protein_id)
                        )
                        if geneInfo["type"][i] == "mRNA":
                            tbl.write("\t\t\tprotein_id\tgnl|ncbi|%s\n" % (protein_id))
                            for num, cds in enumerate(geneInfo["CDS"][i]):
                                # single exon, so slightly differnt method
                                if num == 0 and num == len(geneInfo["CDS"][i]) - 1:
                                    tbl.write(
                                        "%s%s\t%s%s\tCDS\n" % (ps, cds[0], pss, cds[1])
                                    )
                                elif num == 0:
                                    tbl.write("%s%s\t%s\tCDS\n" % (ps, cds[0], cds[1]))
                                # this is last one
                                elif num == len(geneInfo["CDS"][i]) - 1:
                                    tbl.write("%s\t%s%s\n" % (cds[0], pss, cds[1]))
                                else:
                                    tbl.write("%s\t%s\n" % (cds[0], cds[1]))
                            tbl.write(
                                "\t\t\tcodon_start\t%i\n" % geneInfo["codon_start"][i]
                            )
                            if annotations:  # write functional annotation
                                if geneInfo["EC_number"][i]:
                                    for EC in geneInfo["EC_number"][i]:
                                        tbl.write("\t\t\tEC_number\t%s\n" % EC)
                                if geneInfo["db_xref"][i]:
                                    for xref in geneInfo["db_xref"][i]:
                                        tbl.write("\t\t\tdb_xref\t%s\n" % xref)
                                if geneInfo["go_terms"][i]:
                                    for go in geneInfo["go_terms"][i]:
                                        goLine = _goFormat(go)
                                        if goLine:
                                            tbl.write("{:}\n".format(goLine))
                                if geneInfo["note"][i]:
                                    for item in geneInfo["note"][i]:
                                        tbl.write("\t\t\tnote\t%s\n" % item)
                            tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
                            tbl.write(
                                "\t\t\ttranscript_id\tgnl|ncbi|%s_mrna\n" % (protein_id)
                            )
                            tbl.write("\t\t\tprotein_id\tgnl|ncbi|%s\n" % (protein_id))
                    else:  # means this is on crick strand
                        for num, exon in enumerate(geneInfo["mRNA"][i]):
                            # single exon, so slightly differnt method
                            if num == 0 and num == len(geneInfo["mRNA"][i]) - 1:
                                tbl.write(
                                    "%s%s\t%s%s\t%s\n"
                                    % (ps, exon[1], pss, exon[0], geneInfo["type"][i])
                                )
                            elif num == 0:
                                tbl.write(
                                    "%s%s\t%s\t%s\n"
                                    % (ps, exon[1], exon[0], geneInfo["type"][i])
                                )
                            # this is last one
                            elif num == len(geneInfo["mRNA"][i]) - 1:
                                tbl.write("%s\t%s%s\n" % (exon[1], pss, exon[0]))
                            else:
                                tbl.write("%s\t%s\n" % (exon[1], exon[0]))
                        tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
                        tbl.write(
                            "\t\t\ttranscript_id\tgnl|ncbi|%s_mrna\n" % (protein_id)
                        )
                        if geneInfo["type"][i] == "mRNA":
                            tbl.write("\t\t\tprotein_id\tgnl|ncbi|%s\n" % (protein_id))
                            for num, cds in enumerate(geneInfo["CDS"][i]):
                                # single exon, so slightly differnt method
                                if num == 0 and num == len(geneInfo["CDS"][i]) - 1:
                                    tbl.write(
                                        "%s%s\t%s%s\tCDS\n" % (ps, cds[1], pss, cds[0])
                                    )
                                elif num == 0:
                                    tbl.write("%s%s\t%s\tCDS\n" % (ps, cds[1], cds[0]))
                                # this is last one
                                elif num == (len(geneInfo["CDS"][i]) - 1):
                                    tbl.write("%s\t%s%s\n" % (cds[1], pss, cds[0]))
                                else:
                                    tbl.write("%s\t%s\n" % (cds[1], cds[0]))
                            tbl.write(
                                "\t\t\tcodon_start\t%i\n" % geneInfo["codon_start"][i]
                            )
                            if annotations:  # write functional annotation
                                if geneInfo["EC_number"][i]:
                                    for EC in geneInfo["EC_number"][i]:
                                        tbl.write("\t\t\tEC_number\t%s\n" % EC)
                                if geneInfo["db_xref"][i]:
                                    for xref in geneInfo["db_xref"][i]:
                                        tbl.write("\t\t\tdb_xref\t%s\n" % xref)
                                if geneInfo["go_terms"][i]:
                                    for go in geneInfo["go_terms"][i]:
                                        goLine = _goFormat(go)
                                        if goLine:
                                            tbl.write("{:}\n".format(goLine))
                                if geneInfo["note"][i]:
                                    for item in geneInfo["note"][i]:
                                        tbl.write("\t\t\tnote\t%s\n" % item)
                            tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
                            tbl.write(
                                "\t\t\ttranscript_id\tgnl|ncbi|%s_mrna\n" % (protein_id)
                            )
                            tbl.write("\t\t\tprotein_id\tgnl|ncbi|%s\n" % (protein_id))
                elif geneInfo["type"][i] == "tRNA":
                    if geneInfo["strand"] == "+":
                        for num, exon in enumerate(geneInfo["mRNA"][i]):
                            if num == 0:
                                tbl.write(
                                    "%s\t%s\t%s\n"
                                    % (exon[0], exon[1], geneInfo["type"][i])
                                )
                            else:
                                tbl.write("%s\t%s\n" % (exon[0], exon[1]))
                        tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
                        if geneInfo["product"] == "tRNA-Xxx":
                            tbl.write("\t\t\tpseudo\n")
                    else:
                        for num, exon in enumerate(geneInfo["mRNA"][i]):
                            if num == 0:
                                tbl.write(
                                    "%s\t%s\t%s\n"
                                    % (exon[1], exon[0], geneInfo["type"][i])
                                )
                            else:
                                tbl.write("%s\t%s\n" % (exon[1], exon[0]))
                        tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
                        if geneInfo["product"] == "tRNA-Xxx":
                            tbl.write("\t\t\tpseudo\n")
                elif geneInfo["type"][i] in ["rRNA"]:
                    if geneInfo["strand"] == "+":
                        tbl.write(
                            "%s\t%s\t%s\n"
                            % (
                                geneInfo["location"][0],
                                geneInfo["location"][1],
                                geneInfo["type"][i],
                            )
                        )
                        tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
                    else:
                        tbl.write(
                            "%s\t%s\t%s\n"
                            % (
                                geneInfo["location"][1],
                                geneInfo["location"][0],
                                geneInfo["type"][i],
                            )
                        )
                        tbl.write("\t\t\tproduct\t%s\n" % geneInfo["product"][i])
    if output:
        tbl.close()
    return errors, duplicates, pseudo, nocds



import sys
from .fasta import RevComp
from .utils import zopen


'''
# + strand example
OPO1_006208-T1	2883	0	2883	+	scaffold_60	41779	19205	22545	2883	2883	60	NM:i:0	ms:i:2883	AS:i:2623	nn:i:0	ts:A:+	tp:A:P	cm:i:940	s1:i:2816	s2:i:0	de:f:0	rl:i:0	cs:Z::60~gt63ag:365~gt49ag:520~gt53ag:228~gt51ag:560~gt57ag:97~gt67ag:266~gt64ag:660~gt53ag:127
# cs string cs:Z::60~gt63ag:365~gt49ag:520~gt53ag:228~gt51ag:560~gt57ag:97~gt67ag:266~gt64ag:660~gt53ag:127
scaffold_60	funannotate	exon	19206	19265	.	+	.	ID=OPO1_006208-T1.exon1;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	19329	19693	.	+	.	ID=OPO1_006208-T1.exon2;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	19743	20262	.	+	.	ID=OPO1_006208-T1.exon3;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	20316	20543	.	+	.	ID=OPO1_006208-T1.exon4;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	20595	21154	.	+	.	ID=OPO1_006208-T1.exon5;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	21212	21308	.	+	.	ID=OPO1_006208-T1.exon6;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	21376	21641	.	+	.	ID=OPO1_006208-T1.exon7;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	21706	22365	.	+	.	ID=OPO1_006208-T1.exon8;Parent=OPO1_006208-T1;
scaffold_60	funannotate	exon	22419	22545	.	+	.	ID=OPO1_006208-T1.exon9;Parent=OPO1_006208-T1;

# - strand example
OPO1_006778-T1	3150	0	3150	-	scaffold_82	14472	8938	12444	3150	3152	60	NM:i:2	ms:i:3146	AS:i:2910	nn:i:0	ts:A:+	tp:A:P	cm:i:1015	s1:i:3101	s2:i:0	de:f:0.0003	rl:i:0	cs:Z::377~ct45ac:171~ct46ac:72-ct~ct42ac:546~ct60ac:381~ct64ac:695~ct54ac:830~ct43ac:78
# cs string: cs:Z::377~ct45ac:171~ct46ac:72-ct~ct42ac:546~ct60ac:381~ct64ac:695~ct54ac:830~ct43ac:78
scaffold_82	funannotate	exon	12367	12444	.	-	.	ID=OPO1_006778-T1.exon1;Parent=OPO1_006778-T1;
scaffold_82	funannotate	exon	11494	12323	.	-	.	ID=OPO1_006778-T1.exon2;Parent=OPO1_006778-T1;
scaffold_82	funannotate	exon	10745	11439	.	-	.	ID=OPO1_006778-T1.exon3;Parent=OPO1_006778-T1;
scaffold_82	funannotate	exon	10300	10680	.	-	.	ID=OPO1_006778-T1.exon4;Parent=OPO1_006778-T1;
scaffold_82	funannotate	exon	9694	10239	.	-	.	ID=OPO1_006778-T1.exon5;Parent=OPO1_006778-T1;
scaffold_82	funannotate	exon	9578	9649	.	-	.	ID=OPO1_006778-T1.exon6;Parent=OPO1_006778-T1;
scaffold_82	funannotate	exon	9361	9531	.	-	.	ID=OPO1_006778-T1.exon7;Parent=OPO1_006778-T1;
scaffold_82	funannotate	exon	8939	9315	.	-	.	ID=OPO1_006778-T1.exon8;Parent=OPO1_006778-T1;
'''

def paf2dict(paf, fasta, annotation={}, min_mapq=2):
    # minimap2 paf output needs to be run with --cs flag
    with zopen(paf) as infile:
        for line in infile:
            line = line.rstrip()
            cols = line.split('\t')
            qname, qlen, qstart, qend, strand, tname, tlen, tstart, tend, matches, alen, mapq = cols[:12]
            if int(mapq) <= min_mapq:
                continue
            tags = cols[12:]
            cs = None
            for tag in tags:
                if tag.startswith('cs:Z:'):
                    cs = tag
            if not cs:
                sys.stderr.write('Error: {} file does not contain cs:Z: tags, run minimap2 with the -x splice --cs flags\n'.format(paf))
                return {}
            ref_coords, query_coords, properSplice = cs2coords(int(tstart), qstart, int(qlen), strand, cs)
            annotation[qname] = (ref_coords, query_coords, properSplice, cs)
    return annotation


def cs2coords(start, qstart, length, strand, cs, offset=1,
              splice_donor=['gt', 'at'], splice_acceptor=['ag', 'ac']):
    """
    # From minimap2 manual this is the cs flag definitions
    Op	Regex	Description
    =	[ACGTN]+	Identical sequence (long form)
    :	[0-9]+	Identical sequence length
    *	[acgtn][acgtn]	Substitution: ref to query
    +	[acgtn]+	Insertion to the reference
    -	[acgtn]+	Deletion from the reference
    ~	[acgtn]{2}[0-9]+[acgtn]{2}	Intron length and splice signal
    """
    cs = cs.replace('cs:Z:', '')
    ProperSplice = True
    exons = [int(start)]
    position = int(start)
    query = [int(qstart)]
    querypos = 0
    num_exons = 1
    gaps = 0
    indels = []
    if strand == '+':
        sp_donor = splice_donor
        sp_acceptor = splice_acceptor
        sort_orientation = False
    elif strand == '-':
        # rev comp and swap donor/acceptor
        sp_donor = [RevComp(x).lower() for x in splice_acceptor]
        sp_acceptor = [RevComp(x).lower() for x in splice_donor]
        sort_orientation = True
    for s, value in cs2tuples(cs):
        if s == ':':
            position += int(value)
            querypos += int(value)
            indels.append(0)
        elif s == '-':
            gaps += 1
            position += len(value)
            querypos += len(value)
            indels.append(-len(value))
        elif s == '+':
            gaps += 1
            position += len(value)
            querypos += len(value)
            indels.append(len(value))
        elif s == '~':
            if value.startswith(tuple(sp_donor)) and value.endswith(tuple(sp_acceptor)):
                ProperSplice = True
            else:
                ProperSplice = False
            num_exons += 1
            exons.append(position+indels[-1])
            query.append(querypos)
            query.append(querypos+1)
            intronLen = int(value[2:-2])
            position += intronLen
            exons.append(position)
            indels.append(0)
    # add last Position
    exons.append(position)
    query.append(int(length))
    # convert exon list into list of exon tuples
    exontmp = list(zip(exons[0::2], exons[1::2]))
    queryList = list(zip(query[0::2], query[1::2]))
    exonList = []
    for x in sorted(exontmp, key=lambda tup: tup[0], reverse=sort_orientation):
        exonList.append((x[0]+offset, x[1]))
    return exonList, queryList, ProperSplice


def cs2tuples(cs, separators=[':', '*', '+', '-', '~']):
    # separators is an array of strings that are being used to split the the string.
    tmpList = []
    i = 0
    while i < len(cs):
        theSeparator = ""
        for current in separators:
            if current == cs[i:i+len(current)]:
                theSeparator = current
        if theSeparator != "":
            tmpList += [theSeparator]
            i = i + len(theSeparator)
        else:
            if tmpList == []:
                tmpList = [""]
            if(tmpList[-1] in separators):
                tmpList += [""]
            tmpList[-1] += cs[i]
            i += 1
    tupList = list(zip(tmpList[::2], tmpList[1::2]))
    return tupList



import sys
from natsort import natsorted
from .utils import zopen


def sort(args):
    sortGFF3(args.gff3, output=args.out)


def sortGFF3(input, output=False):
    # function to sort GFF3 file but maintain gene, mrna, exon, cds order
    data = []
    features = set()
    comments = []
    with zopen(input) as infile:
        for line in infile:
            if line.startswith("\n"):
                continue
            if line.startswith("#"):
                comments.append(line)
                continue
            line = line.rstrip()
            cols = line.split("\t")
            data.append(cols)
            features.add(cols[2])
    # build sort order dictionary for features
    order_map = {
        "gene": 0,
        "mRNA": 1,
        "transcript": 2,
        "tRNA": 3,
        "ncRNA": 4,
        "rRNA": 5,
        "pseudogene": 6,
        "five_prime_utr": 7,
        "five_prime_UTR": 8,
        "exon": 9,
        "CDS": 10,
        "three_prime_utr": 11,
        "three_prime_UTR": 12,
    }
    idx = len(order_map)
    for x in features:
        if x not in order_map:
            order_map[x] = idx
            idx += 1
    # we can now sort
    sort_data = natsorted(data, key=lambda x: (x[0], int(x[3]), order_map[x[2]]))
    # now we can write back out to file
    if output:
        outfile = zopen(output, mode="w")
    else:
        outfile = sys.stdout
    for y in comments:
        outfile.write(y)
    for x in sort_data:
        outfile.write("{}\n".format("\t".join(x)))
    if output:
        outfile.close()



from .utils import zopen, readBlocks
import os


def go_term_dict():
    """Parses the Gene Ontology file and returns GO lookup dictionary, version, and date.

    The go.obo.gz file is included in the GFFtk package, it can be updated in site-packages/gfftk/data/go.obo.gz

    Returns
    -------
    go : dict of dict
        dictionary of go_term: { "name": str, "namespace": str}
    format_version : str
        version of the go.obo file
    release_date : str
        release data of the go.obo file

    """
    goref = os.path.join(os.path.join(os.path.dirname(__file__)), "data", "go.obo.gz")
    go = {}
    format_version = ""
    release_date = ""
    with zopen(goref) as infile:
        for i, x in enumerate(readBlocks(infile, "[Term]")):
            if i == 0:
                for item in x:
                    if item.startswith("format-version: "):
                        format_version = item.split("format-version: ", 1)[-1].rstrip()
                    elif item.startswith("data-version: releases/"):
                        release_date = item.split("data-version: releases/", 1)[-1].rstrip()
            go_term, go_name, go_space = (None,) * 3
            for item in x:
                if item.startswith("id: GO"):
                    go_term = item.split("id: ", 1)[-1].rstrip()
                elif item.startswith("name: "):
                    go_name = item.split("name: ", 1)[-1].rstrip()
                elif item.startswith("namespace: "):
                    go_space = item.split("namespace: ", 1)[-1].rstrip()
                elif item.startswith("\n"):
                    break
            if go_term and go_name and go_space:
                go[go_term] = {"name": go_name, "namespace": go_space}
    return go, format_version, release_date



from .gff import gff2dict, dict2gff3
from natsort import natsorted
from collections import OrderedDict


def rename(args):
    # load into dictionary
    Genes = gff2dict(args.gff3, args.fasta)

    # now create ordered dictionary and sort by contig and position
    def _sortDict(d):
        return (d[1]["contig"], d[1]["location"][0])

    sGenes = natsorted(iter(Genes.items()), key=_sortDict)
    sortedGenes = OrderedDict(sGenes)
    renamedGenes = {}
    counter = args.numbering
    args.locus_tag = args.locus_tag.rstrip("_")
    transcripts = 0
    for k, v in list(sortedGenes.items()):
        locusTag = args.locus_tag + "_" + str(counter).zfill(6)
        renamedGenes[locusTag] = v
        renamedGenes[locusTag]["gene_synonym"].append(k)
        newIds = []
        for i in range(0, len(v["ids"])):
            newIds.append("{}-T{}".format(locusTag, i + 1))
            transcripts += 1
        renamedGenes[locusTag]["ids"] = newIds
        counter += 1

    # write to gff3
    dict2gff3(renamedGenes, output=args.out)



import platform
import logging
from .__version__ import __version__
import numpy
import natsort
from .interlap import __version__ as interlap_version

GREEN = "\033[92m"
END = "\033[0m"


def startLogging(logfile=False):
    stdoutformat = logging.Formatter(
        GREEN + "%(asctime)s" + END + ": %(message)s", datefmt="[%b %d %I:%M %p]"
    )
    fileformat = logging.Formatter("%(asctime)s: %(message)s", datefmt="[%x %H:%M:%S]")
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    sth = logging.StreamHandler()
    sth.setLevel(logging.INFO)
    sth.setFormatter(stdoutformat)
    logger.addHandler(sth)
    if logfile:
        fhnd = logging.FileHandler(logfile)
        fhnd.setLevel(logging.DEBUG)
        fhnd.setFormatter(fileformat)
        logger.addHandler(fhnd)
    return logger


def system_info(log):
    log(
        "Python v{}; GFFtk v{}; numpy v{}; natsort v{}; interlap v{}".format(
            platform.python_version(),
            __version__,
            numpy.__version__,
            natsort.__version__,
            interlap_version,
        )
    )



"""
`InterLap` does fast interval overlap testing with a simple python data
structure.

It works well on the types of querying done in genomic datasets where we have
10's of thousands of intervals and we check for overlap millions of times. It
is very simple and has no dependencies.

It takes tuples or lists where the first 2 elements are start, end and the
remaining elements can be anything.

>>> from interlap import InterLap
>>> inter = InterLap()

Here create 10K random intervals:

>>> import random
>>> sites = random.sample(range(22, 100000000, 12), 50000)
>>> ranges = [(i, i + random.randint(2000, 20000)) for i in sites]

Add them to the interval tree (this takes < 0.5 seconds):

>>> inter.update(ranges)

We can also add one at a time:

>>> inter.add((20, 22, {'info': 'hi'}))

Now do overlap testing:

>>> [20, 21] in inter
True

>>> next(inter.find((21, 21)))
(20, 22, {'info': 'hi'})

>>> inter.add((2, 3, {'info': 'hello'}))

*NOTE*: below shows how edge-cases are handled.

>>> list(inter.find((2, 2)))
[(2, 3, {'info': 'hello'})]
>>> list(inter.find((3, 3)))
[(2, 3, {'info': 'hello'})]

Test every item in the InterLap. These 50K queries take < 0.5 seconds:

>>> for s, e in ranges:
...     assert (s, e) in inter

>>> for i, se in enumerate(inter):
...     if i > 10: break
...     assert se[0] < se[1]
...

>>> list(inter.closest((2, 2)))
[(2, 3, {'info': 'hello'})]

>>> list(inter.closest((2, 21)))
[(2, 3, {'info': 'hello'}), (20, 22, {'info': 'hi'})]

>>> list(inter.closest((2, 21)))
[(2, 3, {'info': 'hello'}), (20, 22, {'info': 'hi'})]

>>> list(inter.closest((11, 12)))
[(2, 3, {'info': 'hello'}), (20, 22, {'info': 'hi'})]

>>> for i in range(10):
...     inter.add((20, 21))
>>> len(list(inter.closest((10, 13))))
12

>>> for i in range(10):
...     inter.add((18, 21))
>>> len(list(inter.closest((10, 13))))
10

>>> for i in range(10):
...     inter.add((5, 5))
>>> len(list(inter.closest((10, 13))))
20

>>> for i in range(10):
...     inter.add((3, 5))
>>> len(list(inter.closest((10, 13))))
30

>>> inter.add((9, 9))
>>> list(inter.closest((10, 13)))
[(9, 9)]

"""
from itertools import groupby
from operator import itemgetter

__all__ = ['InterLap']

__version__ = '0.2.6'

try:
    int_types = (int, int)
except NameError:
    int_types = (int,)

######################################


def binsearch_left_start(intervals, x, lo, hi):
    while lo < hi:
        mid = (lo + hi)//2
        f = intervals[mid]
        if f[0] < x:
            lo = mid + 1
        else:
            hi = mid
    return lo

# like python's bisect_right find the _highest_ index where the value x
# could be inserted to maintain order in the list intervals


def binsearch_right_end(intervals, x, lo, hi):
    while lo < hi:
        mid = (lo + hi)//2
        f = intervals[mid]
        if x < f[0]:
            hi = mid
        else:
            lo = mid + 1
    return lo


class InterLap(object):

    """Create an Interlap object. (See module docstring)."""

    def __init__(self, ranges=()):
        """The ranges are tuples of (start, end, *whatever)."""
        self._iset = sorted(ranges)
        self._maxlen = max(r[1] - r[0] for r in (ranges or [[0, 0]]))

    def add(self, ranges):
        r"""Add a single (or many) [start, end, \*] item to the tree."""
        if len(ranges) and isinstance(ranges[0], int_types):
            ranges = [ranges]
        iset = self._iset
        self._maxlen = max(self._maxlen, max(r[1] - r[0] + 1 for r in ranges))

        if len(ranges) > 30 or len(iset) < len(ranges):
            iset.extend(ranges)
            iset.sort()
        else:
            for o in ranges:
                iset.insert(binsearch_left_start(iset, o[0], 0, len(iset)), o)

    update = add

    def __len__(self):
        """Return number of intervals."""
        return len(self._iset)

    def find(self, other):
        """Return an interable of elements that overlap other in the tree."""
        iset = self._iset
        l = binsearch_left_start(iset, other[0] - self._maxlen, 0, len(iset))
        r = binsearch_right_end(iset, other[1], 0, len(iset))
        iopts = iset[l:r]
        iiter = (s for s in iopts if s[0] <= other[1] and s[1] >= other[0])
        for o in iiter:
            yield o

    def closest(self, other):
        iset = self._iset
        l = binsearch_left_start(iset, other[0] - self._maxlen, 0, len(iset))
        r = binsearch_right_end(iset, other[1], 0, len(iset))
        l, r = max(0, l - 1), min(len(iset), r + 2)

        while r < len(iset) and iset[r - 1][0] == iset[r][0]:
            r += 1

        while l > 1 and iset[l][1] == iset[l + 1][1]:
            l -= 1
        iopts = iset[l:r]
        ovls = [s for s in iopts if s[0] <= other[1] and s[1] >= other[0]]
        if ovls:
            for o in ovls:
                yield o
        else:
            iopts = sorted(
                [(min(abs(i[0] - other[1]), abs(other[0] - i[1])), i) for i in iopts])
            for dist, g in groupby(iopts, itemgetter(0)):
                # only yield the closest intervals
                for d, ival in g:
                    yield ival
                break

    def __contains__(self, other):
        """Indicate whether `other` overlaps any elements in the tree."""
        iset = self._iset
        l = binsearch_left_start(iset, other[0] - self._maxlen, 0, len(iset))
        # since often the found interval will overlap, we short cut that
        # case of speed.
        max_search = 8
        if l == len(iset):
            return False
        for left in iset[l:l + max_search]:
            if left[1] >= other[0] and left[0] <= other[1]:
                return True
            if left[0] > other[1]:
                return False

        r = binsearch_right_end(iset, other[1], 0, len(iset))
        return any(s[0] <= other[1] and s[1] >= other[0]
                   for s in iset[l + max_search:r])

    def __iter__(self):
        return iter(self._iset)


def overlaps(s1, e1, s2, e2):
    """
    >>> overlaps(2, 4, 3, 5)
    True
    >>> overlaps(2, 4, 4, 5)
    False
    >>> overlaps(2, 200, 3, 5)
    True
    >>> overlaps(3, 5, 2, 200)
    True
    >>> overlaps (3, 5, 5, 6)
    False
    >>> overlaps (5, 6, 3, 5)
    False
    >>> overlaps(2, 4, 2, 4)
    True
    """
    return not (e1 <= s2 or s1 >= e2)


def reduce(args):
    """
    >>> reduce([(2, 4), (4, 9)])
    [(2, 4), (4, 9)]

    >>> reduce([(2, 6), (4, 10)])
    [(2, 10)]
    """
    if len(args) < 2:
        return args
    args.sort()
    ret = [args[0]]
    for next_i, (s, e) in enumerate(args, start=1):
        if next_i == len(args):
            ret[-1] = ret[-1][0], max(ret[-1][1], e)
            break

        ns, ne = args[next_i]
        if e > ns or ret[-1][1] > ns:
            ret[-1] = ret[-1][0], max(e, ne, ret[-1][1])
        else:
            ret.append((ns, ne))
    return ret


class Interval(object):
    """
    >>> i = Interval([(2, 10), (8, 20), (30, 40)])
    >>> i
    Interval([(2, 20), (30, 40)])

    >>> i.add([(20, 22)])
    >>> i
    Interval([(2, 20), (20, 22), (30, 40)])

    >>> i.add([(10, 31)])
    >>> i
    Interval([(2, 40)])

    >>> i.add([(55, 65), (65, 75), (75, 85), (85, 95), (95, 100)])
    >>> i
    Interval([(2, 40), (55, 65), (65, 75), (75, 85), (85, 95), (95, 100)])

    >>> i.add([(1, 95)])
    >>> i
    Interval([(1, 95), (95, 100)])

    >>> i.add(Interval([(90, 100)]))
    >>> i
    Interval([(1, 100)])

    >>> Interval()
    Interval([])

    """

    __slots__ = ('_vals', '_fixed')

    def __init__(self, args=None):
        self._vals = []
        if args is None:
            return
        assert isinstance(args, list)
        if len(args) > 0:
            assert isinstance(args[0], tuple), (args)
            assert isinstance(args[0][0], int)
            self._vals = reduce(args)

    def _as_tuples(self, args):
        vals = []
        if isinstance(args, Interval):
            vals.extend(args._vals)
        else:
            for a in args:
                if isinstance(a, Interval):
                    vals.extend(a._vals)
                else:
                    vals.append(a)
        return vals

    def add(self, args):
        self._vals = reduce(self._vals + self._as_tuples(args))

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self._vals)


if __name__ == "__main__":
    import time
    t0 = time.time()
    import doctest
    print((doctest.testmod(verbose=0, optionflags=doctest.REPORT_ONLY_FIRST_FAILURE)))
    print((time.time() - t0))
